'''
Demonstrates advanced techniques for counting models.

The idea is to decompose a lattice into levels. By going through all
compositions we will create a partitioning of a problem space, as
it can be shown that each lattice can belong to only one composition.

The structure of each composition will allow us to introduce additional
constraints. As a result we will be able to kill more variables, and by
doing so to count lattices on larger domains.

@author: Aleksandar
'''
from logic import bool
from logic.abc import p, i, j, k, l
from logic.fo.quantifiers import *
from logic.cl.runner2 import GPURunner, CPURunner

# increase recursion limit, which is 1000 by default
import sys
sys.setrecursionlimit(5000)


def decompose(s):
    '''Generates all compositions of a given sequence s.'''
    s = tuple(s)
    yield (s,)
    for i in range(1, len(s)):
        for c in decompose(s[i:]):
            yield (s[:i],) + c

def binomial(n, k):
    '''Compute binomial nCk by direct multiplicative method.'''
    if k > n:
        return 0
    if k > n - k:   # Take advantage of symmetry of Pascal's triangle
        k = n - k
    acc = 1
    for i in range(1, k + 1):
        acc *= (n - (k - i))
        acc //= i
    return acc


def assume_reflexive(s, assumptions):
    for i in s:
        assumptions[p(i, i)] = 1

def assume_min_max(s, assumptions):
    '''Fix min and max element.'''
    min_i, *rest, max_i = s
    for i in rest + [max_i]:
        assumptions[p(i, min_i)] = 0
        assumptions[p(min_i, i)] = 1
    for i in [min_i] + rest:
        assumptions[p(max_i, i)] = 0
        assumptions[p(i, max_i)] = 1
    return rest

def assume_unrelated(s, assumptions):
    '''Elements on the same level are unrelated.'''
    for i, j in perm(s, 2):
        assumptions[p(i, j)] = 0

def assume_not_in_relation(s, r, assumptions):
    '''Elements on s level cannot be smaller from elements on lower levels.'''
    for i, j in prod(s, r):
        assumptions[p(i, j)] = 0


def handle_unary_components(composition, assumptions):
    '''Handles levels with one element.'''
    for c_i, c in enumerate(composition):
        if len(c) == 1:
            i = c[0]
            for next_c in composition[(c_i + 1):]:
                for j in next_c:
                    assumptions[p(i, j)] = 1


# i < j iff p(i,j)=1

n = 9

S = range(n)

assumptions = {}

assume_reflexive(S, assumptions)

Sr = assume_min_max(S, assumptions)

Dij = comb(Sr, 2)
Dijk= perm(Sr, 3)

# Statement: p is reflexive
#f1 = A[i : S](p(i, i))  # Not needed since we have assumed reflexivity

# Statement: p is antisymmetric
f2 = A[i, j : Dij](~p(i, j) | ~p(j, i))

# Statement: p is transitive
f3 = A[i, j, k : Dijk](~(p(i, j) & p(j, k)) | p(i, k))

# Every two elements have a meet.
f4 = A[i, j : Dij].E[k : S].A[l : S](p(k, i) & p(k, j) & (~(p(l, i) & p(l, j)) | p(l, k)))

f = bool.Expression(f2 & f3 & f4).eval(assumptions).terms


runner = GPURunner()

n_models = 0
n_c_models = 0

for composition in decompose(Sr):
    c_assumptions = assumptions.copy()
    prev_elements = ()
    coeff = 1
    f_c = bool.true
    for c_i, component in enumerate(composition):
        assume_unrelated(component, c_assumptions)
        assume_not_in_relation(component, prev_elements, c_assumptions)
        coeff *= binomial(len(Sr) - len(prev_elements), len(component))
        prev_elements += component
        if c_i:
            # Every element must be greater from some element on a level below
            f_c &= A[j : component].E[i : composition[c_i - 1]](p(i, j))
    handle_unary_components(composition, c_assumptions)
    expression = bool.Expression(f & f_c).eval(c_assumptions)
    if expression == bool.Expression(bool.true):
        # Reduced to tautology ~ 1
        c_models = 1
    else:
        c_models = runner.count(expression)
    n_models += coeff * c_models
    n_c_models += c_models

    print(composition)
    print('vars: {}'.format(len(expression.variables)))
    print('models: {}'.format(c_models))
    print()

print(n * (n - 1) * n_models)
print(n_c_models)
