'''
Solver of systems of Boolean equations

@author: Aleksandar
'''
import argparse
import os

import numpy as np

from logic import bool
from logic.cl.runner import CPURunner, GPURunner


runner = GPURunner


def find_all(expression):
    result = runner().eval(expression)
    n_vars = len(expression.variables)
    for i in np.where(result != 0)[0]:
        for j, b in enumerate(np.binary_repr(result[i], 32)):
            if b == '1':
                yield np.binary_repr(32 * i + j, n_vars)[::-1]

def find_first(expression, n=1):
    for dummy in range(n):
        yield next(find_all(expression))


def main():
    parser = argparse.ArgumentParser(usage='%(prog)s [-h] [--all] [--cpu] [in [out]]',
                                     description='Searches for variable assignments that satisfy a formula.')
    parser.add_argument('input', metavar='in', nargs='?', default='in.txt',
                        help='a file with a formula to satisfy (default: %(default)s)')
    parser.add_argument('output', metavar='out', nargs='?', default='out.txt',
                        help='a file to write found assignments to (default: %(default)s)')
    parser.add_argument('--all', dest='search', action='store_const', const=find_all, default=find_first,
                        help='find all assignments (default: find the first one)')
    parser.add_argument('--cpu', dest='runner', action='store_const', const=CPURunner, default=GPURunner,
                        help='run on CPU (default: run on GPU)')
    
    args = parser.parse_args()
    if not os.path.dirname(args.output):
        args.output = os.path.join(os.path.dirname(args.input), args.output)
    
    input_name = args.input
    output_name = args.output
    search = args.search
    
    global runner
    runner = args.runner
    
    with open(input_name) as input_file:
        env = vars(bool.algebra)
        
        from logic import abc
        env.update(vars(abc))
        
        from logic.fo import quantifiers
        env.update(vars(quantifiers))
        
        decls = {}
        exec(input_file.read(), env, decls)
        # TODO: switch implementation to GAnd
        eqns = [v for v in decls.values() if isinstance(v, bool.algebra.Term)]
        it = iter(eqns)
        terms = next(it)
        for term in it:
            terms = terms & term
        expression = bool.Expression(terms)
        all_vars = expression.variables
        assumptions = decls.get('assumptions', {})
        expression = expression.eval(assumptions)
        reduced_vars = set(all_vars) - set(assumptions) - set(expression.variables)
    
    with open(output_name, 'w') as output_file:
        print('=' * 80, file=output_file)
        print('Assumed ({}): '.format(len(assumptions)) + ', '.join(sorted('{}={}'.format(k, v) for k, v in assumptions.items())), file=output_file)
        print('Reduced ({}): '.format(len(reduced_vars)) + ', '.join(sorted(str(v) for v in reduced_vars)), file=output_file)
        print('Leftover ({}):'.format(len(expression.variables)), file=output_file)
        print(file=output_file)
        
        if expression == bool.Expression(bool.true):
            print('Reduced to tautology ~ 1', file=output_file)
        elif expression == bool.Expression(bool.false):
            print('Reduced to contradiction ~ 0', file=output_file)
        
        column = '{{!s:^{width}}}'
        column_sep = '   '
        row = column_sep.join(column.format(width=len(str(v))) for v in expression.variables)
        row_width = len(row.format(*expression.variables))
        
        # header
        print(row.format(*expression.variables), file=output_file)
        print('-' * row_width, file=output_file)
        
        # data
        for result in search(expression):
            print(row.format(*result), file=output_file)


if __name__ == '__main__':
#     import cProfile as profile
#     profile.run('main()')
    main()
