'''
Sudoku Solver

Usage: sudoku.py <input_file>

@author: Aleksandar
'''
from logic import bool
from logic.core import ast
from logic.abc import p, i, j, k, l
from logic.fo.quantifiers import *


# Sudoku size
n = 3

# cell indices
S = range(n**2)

# cell values
V = {v + 1 for v in S}
Dkl = comb(V, 2)

# we model a Sudoku puzzle in the following way:
#   cell(i,j)=k iff p(i,j,k)=1

# there is a value in every cell
f1 = A[i : S].A[j : S].E[k : V](p(i, j, k))
# each value appears in each row
f2 = A[i : S].A[k : V].E[j : S](p(i, j, k))
# each value appears in each column
f3 = A[j : S].A[k : V].E[i : S](p(i, j, k))

# in one cell there can be only one value
f4 = A[i : S].A[j : S].A[k, l : Dkl](~(p(i, j, k) & p(i, j, l)))

def subsquare(i, j):
    a, b = i // n, j // n
    x = [a * n + c for c in range(n)]
    y = [b * n + c for c in range(n)]
    return prod(x, y)

# each value appears in each subsquare
f5 = bool.true
for sx, sy in prod(range(0, n**2, n), repeat=2):
    f5 &= A[k : V].E[i, j : subsquare(sx, sy)](p(i, j, k))


def set_assumption(i, j, k, assumptions):
    '''
    Fills in a cell(i,j) with k.
    '''
    for ii in S:
        assumptions[p(ii, j, k)] = 0
    for jj in S:
        assumptions[p(i, jj, k)] = 0
    for kk in V:
        assumptions[p(i, j, kk)] = 0
    for ii, jj in subsquare(i, j):
        assumptions[p(ii, jj, k)] = 0
    assumptions[p(i, j, k)] = 1

def load(assumptions):
    '''
    Loads a Sudoku puzzle from an input file into assumptions.
    '''
    import sys
    with open(sys.argv[1]) as f:
        for i, row in enumerate(f):
            for j, k in enumerate(''.join(row.split())):
                if k != '-':
                    set_assumption(i, j, int(k), assumptions)

def pprint(assumptions):
    '''
    Pretty prints a Sudoku puzzle.
    '''
    sudoku = [['-'] * n**2 for i in range(n**2)]
    for i, j, k in [k.indices for k, v in assumptions.items() if v]:
        sudoku[i][j] = k
    
    row_sep = '+-{}-+'.format('-+-'.join(['-'.join(['-'] * n)] * n))
    row_tl = '| {} |'.format(' | '.join([' '.join(['{}'] * n)] * n))
    for i, row in enumerate(sudoku):
        if i % n == 0:
            print(row_sep)
        print(row_tl.format(*row))
    print(row_sep)
    print()


# Since we don't have negative literals in sudoku, we only search for positive
# ones.
#
# Negative literal ~p(i, j, k) would mean that we can't put k at (i, j).
# That is only possible if have already put k somewhere else in i-th row,
# j-th column or (i, j)-th subsquare. But if we had put k somewhere else,
# we have already eliminated all negative literals containing k, hence we
# can't have negative literals.
class Literals(ast.Visitor):
    
    def __call__(self, term):
        literals = set()
        self.visit((term,), literals)
        return literals
    
    def default_visit(self, term, literals):
        for t in term:
            if isinstance(t, (bool.algebra.And, bool.algebra.Var)):
                self.visit(t, literals)
    
    def visit_Var(self, term, literals):
        literals.add(term)
    
literals = Literals()


def best_guess(variables):
    '''
    Simple strategy for choosing a cell with least options.
    '''
    sudoku = [[[] for i in range(n**2)] for j in range(n**2)]
    for i, j, k in [var.indices for var in variables]:
        sudoku[i][j].append(k)
    min_i, min_j, unused = min([(i, j, len(vals)) for i, row in enumerate(sudoku) for j, vals in enumerate(row) if vals], key=itemgetter(2))
    return [(min_i, min_j, k) for k in sudoku[min_i][min_j]]


def ssolve(f, assumptions):
    '''
    Simple backtracking Sudoku solver.
    '''
    f = f.eval(assumptions)
    while literals(f.terms):
        for i, j, k in [literal.indices for literal in literals(f.terms)]:
            set_assumption(i, j, k, assumptions)
        f = f.eval(assumptions)
    # print(len(f.variables))
    if f == bool.Expression(bool.false):
        return None
    elif f == bool.Expression(bool.true):
        return assumptions
    elif len(f.variables) > 30:
        for i, j, k in best_guess(f.variables):
            assumptions_copy = assumptions.copy()
            set_assumption(i, j, k, assumptions_copy)
            solution = ssolve(f, assumptions_copy)
            if solution:
                return solution
    else:
        print('This could make use of gpu acceleration...')
        return assumptions


def main():
    assumptions = {}
    load(assumptions)
    pprint(assumptions)
    
    f = bool.Expression(f1 & f2 & f3 & f4 & f5)
    solution = ssolve(f, assumptions)
    pprint(solution)


if __name__ == '__main__':
    main()
