'''
Algebra Module

@author: Aleksandar
'''
from abc import abstractstaticmethod

from logic.core.ast import Node, Visitor


class Term(Node):
    '''
    Base class for Boolean algebra terms.
    '''
    
    class __Variables(Visitor):
        '''
        Collects term's variables.
        '''
        
        def __call__(self, term):
            all_vars = set()
            self.visit(term, all_vars)
        
            from operator import attrgetter
            return sorted(all_vars, key=attrgetter(*Var.__slots__))
        
        def visit_Var(self, term, all_vars):
            all_vars.add(term)
    
    # term's variables
    variables = property(__Variables())
    
    def __init__(self, *terms, **kwargs):
        if len(terms) != self.arity():
            raise TypeError('{} expected {} term(s), got {}'.format(self.__class__.__name__, self.arity(), len(terms)))
        super().__init__(*terms, **kwargs)
    
    def __and__(self, other):
        return And(self, other)
    
    def __rand__(self, other):
        return And(other, self)
    
    def __xor__(self, other):
        return Xor(self, other)
    
    def __rxor__(self, other):
        return Xor(other, self)
    
    def __or__(self, other):
        return Or(self, other)
    
    def __ror__(self, other):
        return Or(other, self)
    
    def __invert__(self):
        return Not(self)
    
    @abstractstaticmethod
    def arity(): pass  # @NoSelf


class NullaryTerm(Term):
    
    @staticmethod
    def arity():
        return 0


class UnaryOperator(Term):
    
    @staticmethod
    def arity():
        return 1


class BinaryOperator(Term):
    
    @staticmethod
    def arity():
        return 2


class Var(NullaryTerm):
    
    __slots__ = 'name', 'indices'
    
    def __init__(self, name, *indices, **kwargs):
        self.name = name
        self.indices = indices
        super().__init__(**kwargs)
    
    def __call__(self, *indices):
        return Var(self.name, *indices)
    
    def __hash__(self):
        return hash((self.name, self.indices, super().__hash__()))
    
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.name == other.name and self.indices == other.indices and super().__eq__(other)
        return NotImplemented
    
    def __str__(self):
        return str(self.name) + ''.join(str(i) for i in self.indices)
    
    def __repr__(self):
        return ''.join([super().__repr__()[:-1], repr(self.name), ',' if self.indices else '', ','.join(repr(i) for i in self.indices), ')'])


class Const(NullaryTerm):
    
    __slots__ = 'value'
    
    def __init__(self, value, **kwargs):
        self.value = value
        super().__init__(**kwargs)
    
    def __hash__(self):
        return hash((self.value, super().__hash__()))
    
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.value == other.value and super().__eq__(other)
        return NotImplemented

    def __repr__(self):
        return ''.join([super().__repr__()[:-1], repr(self.value), ')'])

true = Const(-1)
false = Const(0)


#===============================================================================
# TODO: Reconsider multiple inheritance in AST class hierarchy
# 
# At the moment we kind of support multiple inheritance, but that support
# isn't fully thought out yet. Basically the problem is with argument passing.
# It isn't really decided if each method should remove it's arguments down the
# line or not. Either way has its strengths and weaknesses. Aside from error
# handling, if we remove arguments we can't have two methods that use the same
# arguments easily. And if we don't remove arguments we have a problem with
# __new__ and __init__ methods of object itself. It appears that they don't
# accept any arguments, so we would need some base class just to act as a sink
# for passed arguments.
# 
# As far as __init__ is concerned only Node removes its arguments, while other
# classes lean more towards the other way of passing everything down the line.
# On the other hand, as can be seen below, the problem with __new__ is that
# for now the super is from object so we must call it without any arguments.
# 
# Bottom line is that we need a design that handles this issue in clear and
# consistent way throughout the whole library.
# 
# 
#------------------------------------------------------------------------------
# TODO: Consider refactoring of reduction rules
# 
# This is only an idea, but it could be nice if we could move some of the rules
# to the more specific classes. Something in the line of moving rules with
# constants to the Const classes, etc. Of course the challenge is to make that
# code work with both the operators and the __new__ methods.
# 
# 
#------------------------------------------------------------------------------
# TODO: Consider integration with python types
#
# This integration should be done primarily in terms of bools, but ints and
# even strings could prove to be quite useful, especially in light of automatic
# conversions (but they should be researched further).
# 
# 
#------------------------------------------------------------------------------
# TODO: Consider multiple dispatch or multimethods
# 
# This could be really interesting solution for both reductions as well as
# general AST operations.
# 
# This generic functions could even handle integration with python types and
# provide extension points for integration with other libraries of interest.
#===============================================================================

class Not(UnaryOperator):
    
    def __new__(cls, *terms, **kwargs):
        t = terms[0]
        if t == true:
            return false
        elif t == false:
            return true
        elif isinstance(t, cls):    # ~~a --> a
            return t[0]
        return super().__new__(cls)


class And(BinaryOperator):
    
    def __new__(cls, *terms, **kwargs):
        t0, t1 = terms
        if false in terms:
            return false
        elif t0 == true:
            return t1
        elif t1 == true:
            return t0
        elif t0 == t1:      # a & a --> a
            return t0
        elif t0 == ~t1:     # a & ~a --> false
            return false
        return super().__new__(cls)


class Xor(BinaryOperator):
    
    def __new__(cls, *terms, **kwargs):
        t0, t1 = terms
        if t0 == true:      # a ^ true --> ~a
            return ~t1
        elif t0 == false:   # a ^ false --> a
            return t1
        elif t1 == true:
            return ~t0
        elif t1 == false:
            return t0
        elif t0 == t1:      # a ^ a --> false
            return false
        elif t0 == ~t1:     # a ^ ~a --> true
            return true
        return super().__new__(cls)


class Or(BinaryOperator):
    
    def __new__(cls, *terms, **kwargs):
        t0, t1 = terms
        if true in terms:
            return true
        elif t0 == false:
            return t1
        elif t1 == false:
            return t0
        elif t0 == t1:      # a | a --> a
            return t0
        elif t0 == ~t1:     # a | ~a --> true
            return true
        return super().__new__(cls)


class Imp(BinaryOperator): pass
class Eq(BinaryOperator): pass

operators = (Not, And, Xor, Or, Imp, Eq)
