'''
Expression Module

@author: Aleksandar
'''
from collections import Hashable
import random

from logic.bool.algebra import BinaryOperator, Var, true, false, operators
from logic.bool.printer import Printer, bitwise_symbols, boolean_symbols, \
    math_symbols
from logic.bool.transforms import Identity, ToBitwise, ToBoolean


class TEval(Identity):
    '''
    Partially evaluates a term.
    '''
    
    def visit_Var(self, var, values):
        val = values.get(var)
        if val in {true, 1, 'T', '1'}:   # In python True and 1 are considered equal
            return true
        elif val in {false, 0, 'F', '0'}:    # Similarly, False and 0 are equal too
            return false
        return var


class Expression(Hashable):
    '''
    Represents a term that can be evaluated.
    '''
    
    _printer = Printer()
    _print_symbols = math_symbols
    
    def __init__(self, terms):
        self.__terms = terms
    
    def __hash__(self):
        return hash(self.__terms)
    
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__terms == other.__terms
        return NotImplemented
    
    def __repr__(self):
        return ''.join([self.__class__.__name__, '(', repr(self.__terms), ')'])
    
    def __str__(self):
        return self._printer(self.__terms, self._print_symbols)
    
    def eval(self, values):
        return self.__class__(self.eval.transform(self.__terms, values))
    eval.transform = TEval()
    
    def is_bitwise(self):
        return False
    
    def to_bitwise(self):
        if self.is_bitwise():
            return self
        return BitwiseExpression(self.to_bitwise.transform(self.__terms))
    to_bitwise.transform = ToBitwise()
    
    def is_boolean(self):
        return False
    
    def to_boolean(self):
        if self.is_boolean():
            return self
        return BooleanExpression(self.to_boolean.transform(self.__terms))
    to_boolean.transform = ToBoolean()
    
    @property
    def variables(self):
        return self.__terms.variables
    
    @property
    def terms(self):
        return self.__terms


class BitwiseExpression(Expression):
    
    _print_symbols = bitwise_symbols
    
    def is_bitwise(self):
        return True


class BooleanExpression(Expression):
    
    _print_symbols = boolean_symbols
    
    def is_boolean(self):
        return True


def random_expression(var_count, var_name='p'):
    '''
    Creates a random expression.
    '''
    
    def random_term(expansions, depth=10):
        if depth == 0:
            expansions = [e for e in expansions if isinstance(e, Var)]
        
        term = random.choice(expansions)
        
        if isinstance(term, Var):
            return term
        else:
            return term(*[random_term(expansions, depth - 1) for dummy in range(term.arity())])
    
    p = Var(var_name)
    variables = [p(i) for i in range(1, var_count + 1)]
    expansions = list(operators) + variables * (len(operators) // len(variables) + 1)
    extensions = [o for o in operators if issubclass(o, BinaryOperator)]
    
    terms = random_term(expansions)
    used_variables = set(terms.variables)
    while len(used_variables) != var_count:
        term = random_term(expansions)
        if used_variables.issuperset(term.variables):
            for var in term.variables:
                if var in expansions:
                    expansions.remove(var)
        else:
            used_variables.update(term.variables)
        terms = random.choice(extensions)(terms, term)
    
    return Expression(terms)
