'''
Runner Module

@author: Aleksandar
'''
import math
import textwrap

import numpy as np
import pyopencl as cl

from logic import bool
from logic.cl.util import get_best_cpu, get_best_gpu
from logic.core.ast import Transform


class ClRunner:
    '''
    Base class for OpenCL runnners.
    '''
    
    def __init__(self, device):
        self.__device = device
        self.__template = None
        self.__vector_width = 0
    
    @property
    def device(self):
        return self.__device
    
    @property
    def global_mem_size(self):
        return min(self.__device.global_mem_size, 2 ** self.__device.address_bits)
    
    @property
    def max_mem_alloc_size(self):
        if self.__device.platform.name.startswith('NVIDIA'):
            return self.global_mem_size
        return self.__device.max_mem_alloc_size
    
    @property
    def vector_width_int(self):
        if not self.__vector_width:
            return self.__device.preferred_vector_width_int
        return self.__vector_width
    
    @vector_width_int.setter
    def vector_width(self, value):
        if value in {0, 1, 2, 3, 4, 8, 16}:
            self.__vector_width = value
            self.__template = None
    
    @property
    def _max_const_vars(self):
        vector_width_bits = self.vector_width_int * 32
        return math.floor(math.log(vector_width_bits, 2))
    
    @property
    def _max_mem_vars(self):
        '''
        Estimates the maximum number of variables for which the free vectors
        can fit in the available memory M, taking into account the size of
        memory necessary for the result (C).
        
        Basically, we are estimating solution of the following inequality:
        
                            (x + C) * 2^x = M
        
            M - memory size in bytes
            C - vector size in bytes
        '''
        M = self.max_mem_alloc_size
        C = self.vector_width_int * 4
        
        t = math.log(M, 2)                          # t = log2(M)
        return math.floor(t - math.log(t + C, 2))   # floor(t - log2(t + C))
    
    @property
    def template(self):
        if self.__template is None:
            cl_type = 'int'
            cl_cast = ''
            cl_init = '{{{{0[{}][0]}}}}'
            cl_false = 0
            cl_true = -1
            if self.vector_width_int > 1:
                cl_type += str(self.vector_width_int)
                cl_cast = '({})'.format(cl_type)
                cl_init = ' '.join((cl_cast, '{{{{0[{}]}}}}'))
                cl_false = '{} {}'.format(cl_cast, (cl_false,) * self.vector_width_int)
                cl_true = '{} {}'.format(cl_cast, (cl_true,) * self.vector_width_int)
            
            self.__template = textwrap.dedent('''\
            
            __kernel void f(__global char *m, __global {cl_type} *result) {{{{\n''' +
            '\n'.join('                __local {} c{};'.format(cl_type, i + 1) for i in range(self._max_const_vars)) +
            '''
                
                __local {cl_type} p[2];\n\n''' +
                '\n'.join('                c{} = {};'.format(i + 1, cl_init.format(i)) for i in range(self._max_const_vars)) +
                '''
                
                p[0] = {cl_false};
                p[1] = {cl_true};
                
                int gid = get_global_id(0);
                int i = gid * {{1}};
                
                result[gid] = {{2}};
            }}}}
            
            ''').format(cl_type=cl_type, cl_cast=cl_cast, cl_false=cl_false, cl_true=cl_true)
        return self.__template
    
    def eval(self, expression, profiling=False):
        variables = expression.variables
        if not variables:
            return np.zeros(0, dtype=np.int32)
        elif len(variables) <= self._max_const_vars:
            return PyRunner().eval(expression)
        
        ctx = cl.Context([self.__device])
        
        if profiling:
            queue = cl.CommandQueue(ctx, properties=cl.command_queue_properties.PROFILING_ENABLE)
        else:
            queue = cl.CommandQueue(ctx)
        
        n_vars = len(variables)
        n_const_vars = min(n_vars, self._max_const_vars)
        n_mem_vars = n_vars - n_const_vars
        
        if n_mem_vars > self._max_mem_vars:
            raise RuntimeError('To many variables, try less.')
        
        p = bool.fv(n_mem_vars).T
        result = np.zeros(len(p) * self.vector_width_int, dtype=np.int32)
        
        g_size = (len(p),)
        
        p = p.flatten()
        
        mf = cl.mem_flags
        p_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.USE_HOST_PTR, hostbuf=p)
        result_buf = cl.Buffer(ctx, mf.WRITE_ONLY, result.nbytes)
        
        class ClPrinter(bool.printer.Printer):
            
            def __init__(self, src_vars):
                self.src_vars = src_vars
                super().__init__()
                
            def visit_Var(self, term, *args, **kwargs):
                return self.src_vars[term]
        
        src_vars = ['c{}'.format(i + 1) for i in range(n_const_vars)]
        src_vars += ['p[m[i + {}]]'.format(i - n_const_vars) for i in range(n_const_vars, n_vars)]
        src_vars = dict(zip(variables, src_vars))
        
        expression = expression.to_bitwise()
        expression._printer = ClPrinter(src_vars)
        
        src = self.template.format(tuple(tuple(fv) for fv in bool.fv32(n_const_vars)), n_mem_vars, expression)
        
        program = cl.Program(ctx, src)
        try:
            program.build()
        except:
            print('Build log:')
            print('===============================================================')
            print(program.get_build_info(ctx.devices[0], cl.program_build_info.LOG))
            print('===============================================================')
            raise
        
        try:
            eval_evt = program.f(queue, g_size, None, p_buf, result_buf)
        except:
            import pyopencl.tools as cl_tools
            cl_tools.clear_first_arg_caches()
            raise
        
        if profiling:
            eval_evt.wait()
            elapsed_time = 1e-9 * (eval_evt.profile.end - eval_evt.profile.start)
            #return elapsed_time # XXX: remove me
            print("Evaluation time of expression: %g s" % elapsed_time)
        
        cl.enqueue_copy(queue, result, result_buf).wait()
        
        return result


class CPURunner(ClRunner):
    
    def __init__(self):
        super().__init__(get_best_cpu())
    
    @property
    def vector_width_int(self):
        if self.device.platform.name.startswith('Intel') and not self._ClRunner__vector_width: # XXX: _ClRunner... is a hack
            return 4
        return super().vector_width_int
    
    @property
    def max_mem_alloc_size(self):
        if self.device.platform.name.startswith('Intel'):
            return self.global_mem_size
        return super().max_mem_alloc_size


class GPURunner(ClRunner):
    
    def __init__(self):
        super().__init__(get_best_gpu())


class PyRunner(Transform):
    '''
    PyRunner is intended for expressions with less than 6 variables.
    '''
    
    def eval(self, expression, values=None):
        variables = expression.variables
        if not variables:
            return np.zeros(0, dtype=np.int32)
        
        if not values:
            # TODO: fix free_vectors return order (it's different from fv and fv32)
            values = reversed(bool.free_vectors(len(variables)))
        values = dict(zip(variables, values))
        
        expression = expression.to_bitwise()
        result = self.visit(expression.terms, values)
        
        # Adjusting result to np.int32 which is return type of ClRunners
        if len(variables) < 5:
            result <<= 2**5 - 2**len(variables)
        return np.array([result])
    
    def visit_Const(self, term, values):
        return term.value
    
    def visit_Var(self, term, values):
        return values[term]
    
    def visit_Not(self, term, values):
        op, = self.default_visit(term, values)
        return ~op
    
    def visit_And(self, term, values):
        op1, op2 = self.default_visit(term, values)
        return op1 & op2
    
    def visit_Or(self, term, values):
        op1, op2 = self.default_visit(term, values)
        return op1 | op2
    
    def visit_Xor(self, term, values):
        op1, op2 = self.default_visit(term, values)
        return op1 ^ op2
    
    def default_visit(self, node, *args, **kwargs):
        # TODO: reconsider default visit for transforms
        # without this change PyRunner can't work because
        # of same reasons as printer
        return [self.visit(child, *args, **kwargs) for child in node]
