'''
Runner Module v2
  - Implements counting in OpenCL, so that it can be performed by GPU

@author: Aleksandar
'''
import io
import math
import textwrap

import numpy as np
import pyopencl as cl

from logic import bool
from logic.cl.util import get_best_cpu, get_best_gpu
from logic.core.ast import Transform


class ClRunner:
    '''
    Base class for OpenCL runnners.
    '''
    
    def __init__(self, device):
        self.__device = device
        self.__template = None
    
    @property
    def device(self):
        return self.__device
    
    @property
    def global_mem_size(self):
        return min(self.__device.global_mem_size, 2 ** self.__device.address_bits)
    
    @property
    def max_mem_alloc_size(self):
        # if self.__device.platform.name.startswith('NVIDIA'):  # XXX: this only works with TCC driver
            # return self.global_mem_size
        return self.__device.max_mem_alloc_size
    
    @property
    def vector_width_int(self):
        return self.__device.preferred_vector_width_int
    
    @property
    def _max_const_vars(self):
        vector_width_bits = 32 * self.vector_width_int
        return math.floor(math.log(vector_width_bits, 2))
    
    @property
    def _max_vars(self):
        return math.floor(math.log(8 * self.max_mem_alloc_size, 2)) # TODO: this should be multiplied with 8, 
                                                                    # but clenqueuereadbuffer complains that
                                                                    # it is left without resources
                                                                    #
                                                                    # It appears that the problem was lack of
                                                                    # a TCC driver, see XXX
    
    @property
    def template(self):
        if self.__template is None:
            cl_type = 'int'
            cl_cast = ''
            cl_init = '{{{{0[{}][0]}}}}'
            cl_false = 0
            cl_true = -1
            if self.vector_width_int > 1:
                cl_type += str(self.vector_width_int)
                cl_cast = '({})'.format(cl_type)
                cl_init = ' '.join((cl_cast, '{{{{0[{}]}}}}'))
                cl_false = '{} {}'.format(cl_cast, (cl_false,) * self.vector_width_int)
                cl_true = '{} {}'.format(cl_cast, (cl_true,) * self.vector_width_int)
            
            self.__template = ''
            
            if float(self.__device.version.split()[1]) < 1.2:
                self.__template += textwrap.dedent('''\
                
                /*
                * Hacker's Delight 32 bit pop function:
                * http://www.hackersdelight.org/HDcode/newCode/pop_arrayHS.c.txt
                */
                {cl_type} popcount({cl_type} x) {{{{
                    x = x - ((x >> 1) & 0x55555555);
                    x = (x & 0x33333333) + ((x >> 2) & 0x33333333);
                    x = (x + (x >> 4)) & 0x0F0F0F0F;
                    x = x + (x >> 8);
                    x = x + (x >> 16);
                    return x & 0x0000003F;
                }}}}
                
                ''').format(cl_type=cl_type)
            
            self.__template += textwrap.dedent('''\
            
            {cl_type} f(uint gid, __local {cl_type} *c, __local {cl_type} *p) {{{{
                bool v[{{1}}];
                uint m = 1;
                for (uint j=0; j < {{1}}; j++) {{{{
                    v[j] = gid & m;
                    m <<= 1;
                }}}}
                
                return {{2}};
            }}}}
            
            
            __kernel void eval(__global {cl_type} *result) {{{{
                __local {cl_type} c[{const_vars}];
                
                __local {cl_type} p[2];\n\n''' +
                '\n'.join('                c[{}] = {};'.format(i, cl_init.format(i)) for i in range(self._max_const_vars)) +
                '''
                
                p[0] = {cl_false};
                p[1] = {cl_true};
                
                uint gid = get_global_id(0);
                
                result[gid] = f(gid, c, p);
            }}}}
            
            
            __kernel void zeros(__global {cl_type} *result) {{{{
                uint gid = get_global_id(0);
                
                result[gid] = {cl_false};
            }}}}
            
            
            __kernel void count(__global {cl_type} *result) {{{{
                __local {cl_type} c[{const_vars}];
                
                __local {cl_type} p[2];\n\n''' +
                '\n'.join('                c[{}] = {};'.format(i, cl_init.format(i)) for i in range(self._max_const_vars)) +
                '''
                
                p[0] = {cl_false};
                p[1] = {cl_true};

                uint gid = get_global_id(0);
                
                result[gid] += ({cl_type}) ({{3}}) * popcount(f(gid, c, p));
            }}}}
            
            
            __kernel void init_findmodel(__global uint *result) {{{{
                result[0] = 0;
            }}}}
            
            
            __kernel void findmodel(__global uint *result) {{{{
                __local {cl_type} c[{const_vars}];
                
                __local {cl_type} p[2];\n\n''' +
                '\n'.join('                c[{}] = {};'.format(i, cl_init.format(i)) for i in range(self._max_const_vars)) +
                '''
                
                p[0] = {cl_false};
                p[1] = {cl_true};
                
                if (!result[0]) {{{{
                    uint gid = get_global_id(0);\n\n''' +
                    ('                    if (any(popcount(f(gid, c, p)) != 0)) {{{{' if self.vector_width_int > 1 else '                    if (popcount(f(gid, c, p))) {{{{') +
                    '''
                        result[0] = gid;
                    }}}}
                }}}}
            }}}}
            
            ''').format(cl_type=cl_type, cl_cast=cl_cast, cl_false=cl_false, cl_true=cl_true, const_vars=self._max_const_vars)
        return self.__template
    
    def eval(self, expression, profiling=False):
        variables = expression.variables
        if not variables:
            return np.zeros(0, dtype=np.int32)
        
        n_vars = len(variables)
        if n_vars > self._max_vars:
            raise RuntimeError('To many variables, try less.')
        elif n_vars <= self._max_const_vars:
            return Models(variables, PyRunner().eval(expression))
        
        n_const_vars = min(n_vars, self._max_const_vars)
        
        ctx = cl.Context([self.__device])
        
        if profiling:
            queue = cl.CommandQueue(ctx, properties=cl.command_queue_properties.PROFILING_ENABLE)
        else:
            queue = cl.CommandQueue(ctx)
        
        result = np.zeros(2**(n_vars - 5), dtype=np.int32)
        
        mf = cl.mem_flags
        result_buf = cl.Buffer(ctx, mf.WRITE_ONLY, result.nbytes)
        
        class ClPrinter(bool.printer.Printer):
            
            def __init__(self, src_vars):
                self.src_vars = src_vars
                super().__init__()
                
            def visit_Var(self, term, *args, **kwargs):
                return self.src_vars[term]
        
        src_vars = ['c[{}]'.format(i) for i in range(n_const_vars)]
        src_vars += ['p[v[{}]]'.format(i - n_const_vars) for i in range(n_const_vars, n_vars)]
        src_vars = dict(zip(variables, src_vars))
        
        expression = expression.to_bitwise()
        expression._printer = ClPrinter(src_vars)
        
        src = self.template.format(tuple(tuple(fv) for fv in bool.fv32(n_const_vars)), n_vars - n_const_vars, expression, 0)    # TODO: refactor templates so that we don't need 0,
                                                                                                                                # it is needed because of count's s_coeff parameter
        program = cl.Program(ctx, src)
        try:
            program.build()
        except:
            print('Build log:')
            print('===============================================================')
            print(program.get_build_info(ctx.devices[0], cl.program_build_info.LOG))
            print('===============================================================')
            raise
        
        try:
            g_size = (2**(n_vars - n_const_vars),)
            eval_evt = program.eval(queue, g_size, None, result_buf)
        except:
            import pyopencl.tools as cl_tools
            cl_tools.clear_first_arg_caches()
            raise
        
        if profiling:
            eval_evt.wait()
            elapsed_time = 1e-9 * (eval_evt.profile.end - eval_evt.profile.start)
            print("Evaluation time of expression: %g s" % elapsed_time)
        
        cl.enqueue_copy(queue, result, result_buf).wait()
        
        return Models(variables, result)
    
    def count(self, expression):
        py_count = np.vectorize(lambda x: np.binary_repr(x, 32).count('1'))
        
        variables = expression.variables
        if not variables:
            if expression == bool.Expression(bool.true):
                return 1
            else:
                return 0
        
        # handle possible reductions in bitwise transform
        coeff = 2**len(variables)
        expression = expression.to_bitwise()
        variables = expression.variables
        coeff //= 2**len(variables)
        
        reps = 0
        
        n_vars = len(variables)
        if n_vars > self._max_vars:
            import warnings
            warnings.warn('You have {} variables, so this may take a while.'.format(n_vars))
            
            reps = n_vars - self._max_vars
        elif n_vars <= self._max_const_vars:
            return coeff * int(np.sum(py_count(PyRunner().eval(expression))))
        
        ctx = cl.Context([self.__device])
        
        queue = cl.CommandQueue(ctx)
        
        if reps:
            result = np.zeros(2**(self._max_vars - 5), dtype=np.int32)
        else:
            result = np.zeros(2**(n_vars - 5), dtype=np.int32)
        
        mf = cl.mem_flags
        result_buf = cl.Buffer(ctx, mf.READ_WRITE, result.nbytes)
        result_buf_init = False
        
        class ClPrinter(bool.printer.Printer):
            
            def __init__(self, src_vars):
                self.src_vars = src_vars
                super().__init__()
                
            def visit_Var(self, term, *args, **kwargs):
                return self.src_vars[term]
        
        from itertools import product
        sum_ = 0
        import time
        for rep in product([0, 1], repeat=reps):
            print('Case:', rep)
            c_start = time.time()
            
            sub_expression = expression
            if rep:
                sub_expression = expression.eval(dict(zip(variables[self._max_vars:], rep)))
                print('Reduction... Done', '{:.4f}'.format(time.time() - c_start))
                c_start = time.time()
            
            s_variables = sub_expression.variables
            n_vars = len(s_variables)
            
            s_coeff = coeff
            if rep:
                s_coeff *= 2**(self._max_vars - n_vars)
            
            if not s_variables:
                if sub_expression == bool.Expression(bool.true):
                    sum_ += s_coeff
                continue
            
            if n_vars <= self._max_const_vars:
                sum_ += s_coeff * int(np.sum(py_count(PyRunner().eval(sub_expression))))
                continue
            
            n_const_vars = self._max_const_vars
            
            src_vars = ['c[{}]'.format(i) for i in range(n_const_vars)]
            src_vars += ['p[v[{}]]'.format(i - n_const_vars) for i in range(n_const_vars, n_vars)]
            src_vars = dict(zip(s_variables, src_vars))
            
            sub_expression._printer = ClPrinter(src_vars)
            
            src = self.template.format(tuple(tuple(fv) for fv in bool.fv32(n_const_vars)), n_vars - n_const_vars, sub_expression, s_coeff)
            print('Translation... Done', '{:.4f}'.format(time.time() - c_start))
            c_start = time.time()
            
            program = cl.Program(ctx, src)
            try:
                program.build()
                print('Compilation... Done', '{:.4f}'.format(time.time() - c_start))
            except:
                print('Build log:')
                print('===============================================================')
                print(program.get_build_info(ctx.devices[0], cl.program_build_info.LOG))
                print('===============================================================')
                raise
            
            try:
                g_size = (2**(n_vars - n_const_vars),)
                if not result_buf_init:
                    result_buf_init = True
                    program.zeros(queue, g_size, None, result_buf)
                print('GPU...')
                program.count(queue, g_size, None, result_buf)
            except:
                import pyopencl.tools as cl_tools
                cl_tools.clear_first_arg_caches()
                raise
        
        cl.enqueue_copy(queue, result, result_buf).wait()
        
        return sum_ + int(np.sum(result, dtype=np.int64))
    
    def exist(self, expression):
        if self.findmodel(expression) is not None:
            return True
        return False
    
    def findmodel(self, expression):
        assumptions = dict.fromkeys(expression.variables, 0)
        
        expression = expression.to_bitwise()
        variables = expression.variables
        if not variables:
            if expression == bool.Expression(bool.true):
                return Model(assumptions)
            return None
        
        assumptions = dict.fromkeys(assumptions.keys() - variables, 0)
        
        reps = 0
        
        n_vars = len(variables)
        if n_vars > self._max_vars:
            import warnings
            warnings.warn('You have {} variables, so this may take a while.'.format(n_vars))
            
            reps = n_vars - self._max_vars
        elif n_vars <= self._max_const_vars:
            for model in self.eval(expression):
                model.update(assumptions)
                return model
            return None
        
        ctx = cl.Context([self.__device])
        
        queue = cl.CommandQueue(ctx)
        
        result = np.zeros(1, dtype=np.uint32)
        
        mf = cl.mem_flags
        result_buf = cl.Buffer(ctx, mf.READ_WRITE, result.nbytes)
        result_buf_init = False
        
        class ClPrinter(bool.printer.Printer):
            
            def __init__(self, src_vars):
                self.src_vars = src_vars
                super().__init__()
                
            def visit_Var(self, term, *args, **kwargs):
                return self.src_vars[term]
        
        from itertools import product
        for rep in product([0, 1], repeat=reps):
            sub_expression = expression
            sub_assumptions = assumptions.copy()
            
            if rep:
                sub_expression = expression.eval(dict(zip(variables[self._max_vars:], rep)))
                sub_assumptions.update(dict.fromkeys(set(variables) - set(sub_expression.variables), 0))
                sub_assumptions.update(zip(variables[self._max_vars:], rep))
            
            s_variables = sub_expression.variables
            n_vars = len(s_variables)
            
            if not s_variables:
                if sub_expression == bool.Expression(bool.true):
                    return Model(sub_assumptions)
                continue
            
            if n_vars <= self._max_const_vars:
                for model in self.eval(sub_expression):
                    model.update(sub_assumptions)
                    return model
                continue
            
            n_const_vars = self._max_const_vars
            
            src_vars = ['c[{}]'.format(i) for i in range(n_const_vars)]
            src_vars += ['p[v[{}]]'.format(i - n_const_vars) for i in range(n_const_vars, n_vars)]
            src_vars = dict(zip(s_variables, src_vars))
            
            sub_expression._printer = ClPrinter(src_vars)
            
            src = self.template.format(tuple(tuple(fv) for fv in bool.fv32(n_const_vars)), n_vars - n_const_vars, sub_expression, 0)    # TODO: refactor templates so that we don't need 0,
                                                                                                                                        # it is needed because of count's s_coeff parameter
            program = cl.Program(ctx, src)
            try:
                program.build()
            except:
                print('Build log:')
                print('===============================================================')
                print(program.get_build_info(ctx.devices[0], cl.program_build_info.LOG))
                print('===============================================================')
                raise
            
            try:
                g_size = (2**(n_vars - n_const_vars),)
                if not result_buf_init:
                    result_buf_init = True
                    program.init_findmodel(queue, (1,), None, result_buf)
                program.findmodel(queue, g_size, None, result_buf)
                cl.enqueue_copy(queue, result, result_buf).wait()
            except:
                import pyopencl.tools as cl_tools
                cl_tools.clear_first_arg_caches()
                raise
            
            for model in self.eval(sub_expression.eval(dict(zip(s_variables[n_const_vars:], np.binary_repr(result[0], 32)[::-1])))):
                model.update(dict.fromkeys(set(s_variables[:n_const_vars]) - model.keys(), 0))
                model.update(zip(s_variables[n_const_vars:], map(int, np.binary_repr(result[0], 32)[::-1])))
                model.update(sub_assumptions)
                return model
        
        return None


class CPURunner(ClRunner):
    
    def __init__(self):
        super().__init__(get_best_cpu())
    
    @property
    def vector_width_int(self):
        if self.device.platform.name.startswith('Intel'):
            return 4
        return super().vector_width_int
    
    @property
    def max_mem_alloc_size(self):
        # if self.device.platform.name.startswith('Intel'): # TODO: reconsider
            # return self.global_mem_size // 2
        return super().max_mem_alloc_size


class GPURunner(ClRunner):
    
    def __init__(self):
        super().__init__(get_best_gpu())


class PyRunner(Transform):
    '''
    PyRunner is intended for expressions with low count of variables, e.g. less than 10.
    '''
    
    def eval(self, expression, values=None):
        variables = expression.variables
        if not variables:
            return np.zeros(0, dtype=np.int32)
        
        n_vars = len(variables)
        
        if not values:
            # TODO: fix free_vectors return order (it's different from fv and fv32)
            values = reversed(bool.free_vectors(n_vars))
        values = dict(zip(variables, values))
        
        expression = expression.to_bitwise()
        result = self.visit(expression.terms, values)
        
        # Adjusting result to np.int32 which is a return type of ClRunners
        r = []
        if n_vars < 5:
            result <<= 2**5 - 2**n_vars
        else:
            for i in range(2**(n_vars - 5) - 1):
                r.append(result % 2**2**5)
                result //= 2**2**5
        r.append(result)
        r.reverse()
        
        return np.array(r).astype(np.int32)
    
    def visit_Const(self, term, values):
        return term.value
    
    def visit_Var(self, term, values):
        return values[term]
    
    def visit_Not(self, term, values):
        op, = self.default_visit(term, values)
        return ~op
    
    def visit_And(self, term, values):
        op1, op2 = self.default_visit(term, values)
        return op1 & op2
    
    def visit_Or(self, term, values):
        op1, op2 = self.default_visit(term, values)
        return op1 | op2
    
    def visit_Xor(self, term, values):
        op1, op2 = self.default_visit(term, values)
        return op1 ^ op2
    
    def default_visit(self, node, *args, **kwargs):
        # TODO: reconsider default visit for transforms
        # without this change PyRunner can't work because
        # of same reasons as printer
        return [self.visit(child, *args, **kwargs) for child in node]


class Models:
    
    def __init__(self, variables, data):
        self.variables = tuple(variables)
        self._raw = data
    
    def _values(self):
        result = self._raw
        n_vars = len(self.variables)
        for i in np.where(result != 0)[0]:
            for j, b in enumerate(np.binary_repr(result[i], 32)):
                if b == '1':
                    yield np.binary_repr(32 * i + j, n_vars)[::-1]
    
    def __len__(self):
        np_count = np.vectorize(lambda x: np.binary_repr(x, 32).count('1'))
        return int(np.sum(np_count(self._raw)))
    
    def __iter__(self):
        m_vars = self.variables
        for value in self._values():
            yield Model(zip(m_vars, map(int, value)))
    
    def __str__(self):
        column = '{{:^{width}}}'
        column_sep = ' '
        row = column_sep.join(column.format(width=len(str(v))) for v in self.variables)
        row_width = len(row.format(*self.variables))
        
        with io.StringIO() as output:
            # header
            print(row.format(*self.variables), file=output)
            print('-' * row_width, file=output)
            
            # data
            for value in self._values():
                print(row.format(*value), file=output)
            
            return output.getvalue()


class Model(dict):
    
    def as_table(self):
        x = sorted({v.indices[0] for v in self})
        y = sorted({v.indices[1] for v in self})
        p = next(iter(self))
        
        column = '{{:^{width}}}'
        hcolumn = column.format(width=max(len(p.name), *map(len, map(str, x))))
        column_sep = ' '
        row = hcolumn + ' | ' + column_sep.join(column.format(width=len(str(i))) for i in y)
        row_width = len(row.format(p.name, *y))
        
        with io.StringIO() as output:
            # header
            print(row.format(p.name, *y), file=output)
            k = row.format(p.name, *y).index('|')
            sep = '-' * row_width
            sep = sep[:k] + '+' + sep[k+1:]
            print(sep, file=output)
            
            # data
            for i in x:
                print(row.format(i, *[self.get(p(i,j), 'x') for j in y]), file=output)
                    
            return output.getvalue()
