'''
AST Module

@author: Aleksandar
'''
from abc import ABCMeta
from collections import Hashable, Sequence


class ASTMeta(ABCMeta):
    '''
    Metaclass for AST classes.
    '''
    
    def __new__(cls, name, bases, namespace):
        '''
        Ensures that __slots__ is present in order to suppress __dict__.
        '''
        slots = namespace.setdefault('__slots__', ())
        if isinstance(slots, str):
            namespace['__slots__'] = (slots,)
        return super().__new__(cls, name, bases, namespace)


class Node(Hashable, Sequence, metaclass=ASTMeta):
    
    __slots__ = '__children', '_hash'
    
    def __init__(self, *nodes, **kwargs):
        if hasattr(self, '_Node__children'):
            # handles the case of reinitialization triggered
            # by the reduction rules performed in subclass's __new__
            #
            # TODO: move to the metaclass. That way it will be handled in the actual subclass.
            return
        self.__children = nodes
        self._hash = hash(self.__children)
        super().__init__(**kwargs)
    
    def __hash__(self):
        return self._hash
    
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__children == other.__children
        return NotImplemented
    
    def __getitem__(self, index):
        return self.__children[index]
    
    def __len__(self):
        return len(self.__children)
    
    def __repr__(self):
        return ''.join([self.__class__.__name__, '(', ','.join(repr(child) for child in self.__children), ')'])
    
    def __getnewargs__(self):
        return self.__children


class Visitor:
    '''
    Base class for AST visitors.
    '''
    
    def __init__(self):
        # TODO: move to __new__. That way the subclasses wont have to call super().__init__
        self._cache = {}
    
    def __call__(self, node, *args, **kwargs):
        return self.visit(node, *args, **kwargs)
    
    def visit(self, node, *args, **kwargs):
        meth = self._cache.get(node.__class__, None)
        if meth is None:
            for cls in node.__class__.__mro__:
                meth = getattr(self, 'visit_' + cls.__name__, None)
                if meth is not None:
                    break
            else:
                meth = self.default_visit
            self._cache[node.__class__] = meth
        return meth(node, *args, **kwargs)
    
    def default_visit(self, node, *args, **kwargs):
        for child in node:
            self.visit(child, *args, **kwargs)


# import traceback

class Transform(Visitor):
    '''
    Base class for AST transforms.
    '''
    
    def default_visit(self, node, *args, **kwargs):
        return node.__class__(*[self.visit(child, *args, **kwargs) for child in node])
#         stack_depth = len(traceback.format_stack())
#         print('-', stack_depth, repr(node))
#         blah = [self.visit(child, *args, **kwargs) for child in node]
#         print('+', stack_depth, node.__class__.__name__, repr(blah))
#         return node.__class__(*blah)
