'''
Quantifiers Module

@author: Aleksandar
'''
from itertools import combinations, permutations, product, groupby
from operator import itemgetter

from logic.bool.algebra import true, false
from logic.bool.transforms import Identity


def reusable(generator):
    '''
    Decorates a generator so that it can be iterated over multiple times.
    '''
    def wrapper(*args, **kwargs):
        class Iterable:
            def __iter__(self):
                return generator(*args, **kwargs)
        return Iterable()
    return wrapper

comb = reusable(combinations)
perm = reusable(permutations)
prod = reusable(product)


class TReindex(Identity):
    '''
    Performs reindexing of variables.
    '''
    
    def visit_Var(self, var, indices, new_indices):
        index_map = dict(zip(indices, new_indices))
        return var(*[index_map.get(i, i) for i in var.indices])


class BigBase:
    '''
    Base class for big operators.

    It builds up AST by iterating index variables over given values.
    '''
    
    _reindex = TReindex()
    
    def __init__(self, expr, index_vars):
        self.expr = expr
        self.index_vars = index_vars
    
    def reindex(self, new_indices):
        return self._reindex(self.expr, self.index_vars, new_indices)
    
    def reindex2(self, indices, new_index):
        # TODO: rename
        return self._reindex(self.expr.over(indices), self.index_vars, new_index)
    
    def over(self, indices):
        if isinstance(self.expr, BigBase):
            new_domain = itemgetter(slice(len(self.index_vars), None))
            return self.big_op([self.reindex2(map(new_domain, d), i) for i, d in groupby(indices, itemgetter(slice(len(self.index_vars))))])
        if len(self.index_vars) == 1:
            # if we got here through reindex2, indices are already tuples
            indices = [i if isinstance(i, tuple) else (i,) for i in indices]
        return self.big_op([self.reindex(i) for i in indices])
    
    def big_op(self, expressions):
        pass


class BigAnd(BigBase):
    
    def big_op(self, expressions):
        # TODO: switch implementation to GAnd
        terms = true
        for term in expressions:
            terms = terms & term
        return terms


class BigOr(BigBase):
    
    def big_op(self, expressions):
        # TODO: switch implementation to GOr
        terms = false
        for term in expressions:
            terms = terms | term
        return terms


def _big_op(big_op, expr, index_vars, indices=None):
    # we use a fact that variables are also nullary terms / empty nodes
    # to check if we got only one index var
    if not index_vars:
        index_vars = (index_vars,)
    if indices is None:
        return big_op(expr, index_vars)
    return big_op(expr, index_vars).over(indices)


def big_and(expr, index_vars, indices=None):
    return _big_op(BigAnd, expr, index_vars, indices)


def big_or(expr, index_vars, indices=None):
    return _big_op(BigOr, expr, index_vars, indices)


class Quantifier:
    '''
    Base class for quantifiers.

    Implements new quantifier syntax: Q[i,j : DxD].Q[k : D](...)
    '''
    
    def __init__(self, index_vars=None, indices=None):
        self.index_vars = index_vars
        self.indices = indices
    
    def __getitem__(self, key):
        if isinstance(key, tuple):
            index_vars = []
            for k in key:
                if isinstance(k, slice):
                    break
                index_vars.append(k)
            index_vars.append(k.start)
            indices = k.stop
        elif isinstance(key, slice):
            index_vars = key.start
            indices = key.stop
        else:
            index_vars = key
            indices = None
        if hasattr(self, 'parent'):
            self.index_vars = index_vars
            self.indices = indices
            return self
        return self.__class__(index_vars, indices)
    
    def __getattr__(self, name):
        if name == 'A':
            a = A.__class__()
            a.parent = self
            return a
        elif name == 'E':
            e = E.__class__()
            e.parent = self
            return e
        raise AttributeError('{!r} object has no attribute {!r}'.format(type(self).__name__, name))
    
    def __call__(self, expr):
        expr = self._big_op(expr, self.index_vars, self.indices)
        while hasattr(self, 'parent'):
            self = self.parent
            expr = self._big_op(expr, self.index_vars, self.indices)
        return expr


class A(Quantifier):
    '''
    Universal quantifier.
    '''
    
    _big_op = lambda *args: big_and(*args[1:])

A = A()


class E(Quantifier):
    '''
    Existential quantifier.
    '''
    
    _big_op = lambda *args: big_or(*args[1:])

E = E()
